==> bordeaux-0.8.1.tar.gz: bordeaux-0.8.1/src/utils.cpp <==
/*
 * Copyright (C) 2020 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <moderncli/print.hpp>
#include <moderncli/sync.hpp>

#include <fstream>
#include <iomanip>
#include <mutex>
#include <cstdio>
#include <cstdarg>

#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <pthread.h>

#include "drivers.hpp"
#include "utils.hpp"

using namespace server;

namespace {
std::ofstream logfile;
std::mutex loglock;

const char alphabet[65] =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

auto u8_1b(const uint8_t *bp) {
    return bp[0] == 0x09 || bp[0] == 0x0a || bp[0] == 0x0d || (bp[0] >= 0x20 && bp[0] <= 0x7e);
}

auto u8_2b(const uint8_t *bp) {
     return bp[0] >= 0xc2 && bp[0] <= 0xdf && bp[1] >= 0x80 && bp[1] <= 0xbf;
}

auto u8_3b(const uint8_t *bp) {
    if(bp[0] == 0xe0 && bp[1] >= 0xa0 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf)
        return true;

    if(bp[0] >= 0xe1 && bp[0] <= 0xef && bp[0] != 0xed && bp[1] >=0x80 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf)
        return true;

    if(bp[0] == 0xed && bp[1] >= 0x80 && bp[1] <= 0x9f && bp[2] >= 0x80 && bp[2] < 0xbf)
        return true;

    return false;
}

auto u8_4b(const uint8_t *bp) {
    if(bp[0] == 0xf0 && bp[1] >= 0x90 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    if(bp[0] >= 0xf1 && bp[0] <= 0xf3 && bp[1] >= 0x80 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    if(bp[0] == 0xf4 && bp[1] >= 0x80 && bp[1] <= 0x8f && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    return false;
}
}

void server::log_reset() {
    const std::lock_guard<std::mutex> lock(loglock);
    if(logfile.is_open())
        logfile.close();
    logfile.open(SERVER_LOGGER_PATH "/bordeaux.log", std::ios_base::app); // FlawFinder: ignore
}

void server::logger(const std::string& msg) {
    if(!logfile.is_open())
        return;

    auto timestamp = system_clock();
    const struct tm now = local_time(timestamp);
    const std::lock_guard<std::mutex> lock(loglock);
    logfile << std::put_time(&now, "%Y-%m-%d %X ") << msg << std::endl;
}

auto util::u8verify(const char *cp, size_t max) -> size_t {
    if(!cp || !*cp || !max)
        return 0;

    size_t size = 0, codepoints = 0;
    auto bp = reinterpret_cast<const uint8_t *>(cp);
    while(*bp && codepoints < max) {
        if(u8_1b(bp)) {
            ++bp;
            ++size;
            ++codepoints;
            continue;
        }

        if(u8_2b(bp)) {
            size += 2;
            bp +=2;
            ++codepoints;
            continue;
        }

        if(u8_3b(bp)) {
            size += 3;
            bp += 3;
            ++codepoints;
            continue;
        }
        if(u8_4b(bp)) {
            size += 4;
            bp += 4;
            ++codepoints;
            continue;
        }

        return 0;
    }

    if(codepoints > max || *bp)
        return 0;

    return size;
}

auto util::compare(std::string_view from, std::string_view to) -> bool {
    auto size = to.size();
    if(!size)
        return false;

    if(from.size() < size)
        return false;

    auto fp = from.data();
    auto tp = to.data();
    while(size--) {
        if((*fp != *tp) && (tolower(*fp) != *tp))
            return false;
        ++fp;
        ++tp;
    }
    return true;
}

auto util::get_value(std::string_view text) -> int {
    int value = 0;
    bool negative = false;

    if(!text.empty() && text.front() == '-') {
        negative = true;
        text.remove_prefix(1);
    }

    for(auto ch : text) {
        if(ch < '0' || ch > '9')
            return 0;
        if(value > 99999999)
            return 0;
        value *= 10;
        value += (ch - '0');
    }

    if(negative)
        return -value;

    return value;
}

auto util::get_env(const char *id, size_t max) -> const char * {
    auto cp = ::getenv(id); // FlawFinder: wrapped in verify
    if(!cp)
        return nullptr;

    if(!u8verify(cp, max))
        return nullptr;

    return cp;
}

// returns milliseconds
auto util::get_duration(const char *cp, long scale) -> long {
    if(!cp)
        return 0;

    auto *fc = strchr(cp, ':');
    if(fc) {
        auto *lc = strrchr(cp, ':');
        auto duration = static_cast<long>(get_value(lc + 1));
        if(fc == lc)
            return ((get_value(cp) * 60L) + duration) * 1000L;
        return ((3600L * get_value(cp)) + (60L * get_value(fc + 1)) + duration) * 1000L;
    }

    char *ep{nullptr};
    auto duration = strtol(cp, &ep, 10);
    if(!ep || !*ep)
        return duration * scale;

    if(!strcmp(ep, "s"))
        return duration * 1000L;

    if(!strcmp(ep, "r"))    // ring timed interval...
        return  duration * Driver::ring_timer() * 1000L;

    if(!strcmp(ep, "m"))
        return duration * 60000L;

    if(!strcmp(ep, "h"))
        return duration * 3600000L;

    if(!strcmp(ep, "ms"))
        return duration;

    return 0;
}

auto util::to_b64(const uint8_t *data, size_t size) -> std::string {
    assert(size > 0);

    std::string result;
    unsigned bits{};

    while(size >= 3) {
        bits = (((unsigned)data[0])<<16) | (((unsigned)data[1])<<8)
            | ((unsigned)data[2]);
        data += 3;
        size -= 3;
        result += format("{}{}{}{}", alphabet[bits >> 18], alphabet[(bits >> 12) & 0x3f], alphabet[(bits >> 6) & 0x3f], alphabet[bits & 0x3f]);
    }

    if(size) {
        bits = ((unsigned)data[0])<<16;
        result += alphabet[bits >> 18];
        if(size == 1) {
            result += alphabet[(bits >> 12) & 0x3f];
            result += '=';
        }
        else {
            bits |= ((unsigned)data[1])<<8;
            result += alphabet[(bits >> 12) & 0x3f];
            result += alphabet[(bits >> 6) & 0x3f];
        }
        result += '=';
    }
    return result;
}

auto util::set_priority(int priority) -> bool {
    auto tid = pthread_self();
    struct sched_param sp{};
    int policy{};

    std::memset(&sp, 0, sizeof(sp));

    policy = SCHED_OTHER;
#ifdef  SCHED_FIFO
    if(priority > 0) {
        policy = SCHED_FIFO;
        priority = sched_get_priority_min(policy) + priority - 1;
        if(priority > sched_get_priority_max(policy))
            priority = sched_get_priority_max(policy);
    }
#endif
#ifdef  SCHED_BATCH
    if(priority < 0) {
        policy = SCHED_BATCH;
        priority = 0;
    }
#endif
    if(policy == SCHED_OTHER)
        priority = 0;

    sp.sched_priority = priority;
    return pthread_setschedparam(tid, policy, &sp) == 0;
}

==> bordeaux-v0.9.2.tar.gz: bordeaux-v0.9.2/src/utils.cpp <==
/*
 * Copyright (C) 2020 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "drivers.hpp"
#include "utils.hpp"

#include <moderncli/datetime.hpp>
#include <moderncli/strings.hpp>
#include <moderncli/socket.hpp>
#include <moderncli/sync.hpp>

#include <fstream>

#include <fcntl.h>
#include <sched.h>

using namespace server;

namespace {
std::ofstream logfile;
std::mutex loglock;
} // end namespace

void server::log_reset() {
    const std::lock_guard<std::mutex> lock(loglock);
    if(logfile.is_open())
        logfile.close();
    logfile.open(SERVER_LOGGER_PATH "/bordeaux.log", std::ios_base::app); // FlawFinder: ignore
}

void server::logger_notify(const std::string& msg, const char *type) {
    if(eq(type, "debug") || !logfile.is_open())
        return;

    auto timestamp = system_clock();
    const struct tm now = local_time(timestamp);
    const std::lock_guard<std::mutex> lock(loglock);
    logfile << std::put_time(&now, ISO_DATETIME) << " " << msg << std::endl;
}

auto util::compare(std::string_view from, std::string_view to) -> bool {
    auto size = to.size();
    if(!size)
        return false;

    if(from.size() < size)
        return false;

    auto fp = from.data();
    auto tp = to.data();
    while(size--) {
        if((*fp != *tp) && (tolower(*fp) != *tp))
            return false;
        ++fp;
        ++tp;
    }
    return true;
}

auto util::get_value(std::string_view text) -> int {
    int value = 0;
    bool negative = false;

    if(!text.empty() && text.front() == '-') {
        negative = true;
        text.remove_prefix(1);
    }

    for(auto ch : text) {
        if(ch < '0' || ch > '9')
            return 0;
        if(value > 99999999)
            return 0;
        value *= 10;
        value += (ch - '0');
    }

    if(negative)
        return -value;

    return value;
}

// returns milliseconds
auto util::get_duration(const char *cp, long scale) -> long {
    if(!cp)
        return 0;

    auto *fc = strchr(cp, ':');
    if(fc) {
        auto *lc = strrchr(cp, ':');
        auto duration = static_cast<long>(get_value(lc + 1));
        if(fc == lc)
            return ((get_value(cp) * 60L) + duration) * 1000L;
        return ((3600L * get_value(cp)) + (60L * get_value(fc + 1)) + duration) * 1000L;
    }

    char *ep{nullptr};
    auto duration = strtol(cp, &ep, 10);
    if(!ep || !*ep)
        return duration * scale;

    if(!strcmp(ep, "s"))
        return duration * 1000L;

    if(!strcmp(ep, "r"))    // ring timed interval...
        return  duration * Driver::ring_timer() * 1000L;

    if(!strcmp(ep, "m"))
        return duration * 60000L;

    if(!strcmp(ep, "h"))
        return duration * 3600000L;

    if(!strcmp(ep, "ms"))
        return duration;

    return 0;
}

auto util::set_priority(int priority) -> bool {
    auto tid = pthread_self();
    struct sched_param sp{};
    int policy{};

    std::memset(&sp, 0, sizeof(sp));

    policy = SCHED_OTHER;
#ifdef  SCHED_FIFO
    if(priority > 0) {
        policy = SCHED_FIFO;
        priority = sched_get_priority_min(policy) + priority - 1;
        if(priority > sched_get_priority_max(policy))
            priority = sched_get_priority_max(policy);
    }
#endif
#ifdef  SCHED_BATCH
    if(priority < 0) {
        policy = SCHED_BATCH;
        priority = 0;
    }
#endif
    if(policy == SCHED_OTHER)
        priority = 0;

    sp.sched_priority = priority;
    return pthread_setschedparam(tid, policy, &sp) == 0;
}

auto util::get_address(const std::string& bind, const std::string& service, int type) -> std::string {
    auto iface(bind);
    if(!ipv6 && (iface.empty() || iface == "*" || iface == "0.0.0.0"))
        return "*";

    if(ipv6 && (iface == "::" || iface == "::*"))
        return "::";

    if(!ipv6 && (iface == "localhost" || iface == "loopback"))
        return "127.0.0.1";

    if(ipv6 && (iface == "localhost" || iface == "loopback"))
        return "::1";


    if(ipv6 && (iface.find_first_of(":") != std::string::npos))
        return iface;

    // find a sip service reference or host lookup...
    if(iface.find_first_of(".") != std::string::npos) {
        Socket::service ifs(iface, service, ipv6 ? AF_INET6 : AF_INET, type);
        if(!ifs)
            ifs.set(iface, "", ipv6 ? AF_INET : AF_INET6, type);
        if(!ifs.empty()) {
            const address_t addr(*ifs);
            return addr.to_string();
        }
    }

    const Socket::interfaces ifa;
    auto addr = ifa.find(bind, ipv6 ? AF_INET6 : AF_INET);
    if(addr)
        iface = inet_host(addr);
    else
        iface = inet_host(bind);

    if(iface.empty())
        throw std::invalid_argument("unknown interface name");
    return iface;
}
